import numpy as np
from .common import *
from . import rotation

def to_homogeneous(x):
    x = np.asarray(x)
    o = np.ones_like(x[..., :1])
    return np.concatenate([x, o], axis=-1)

def from_homogeneous(x):
    return x[..., :-1] / x[..., -1:]

def compose(r, t, rtype, out=None):
    if out is None:
        shape = tuple(np.shape(t)[:-1]) + (4, 4)
        out = np.zeros(shape, dtype=t.dtype)
    rtype.to_matrix(r, out=out[..., :3, :3])
    out[..., :3, 3:] = t.reshape(out[...,:3,3:].shape)
    return out

def translation_from_matrix(T):
    return T[..., :3, 3]

def rotation_from_matrix(T):
    return T[..., :3, :3]

def rotation_2d(x, R=None, c=None, s=None):
    if R is None:
        shape = tuple(np.shape(x)[:-1]) + (2, 2)
        R = np.zeros(shape, dtype=x.dtype)
    if c is None:
        c = np.cos(x)
    if s is None:
        s = np.sin(x)
    R[..., 0, 0] = c
    R[..., 0, 1] = -s
    R[..., 1, 0] = s
    R[..., 1, 1] = c
    return R

def Rz(x, T=None, c=None, s=None):
    if T is None:
        shape = tuple(np.shape(x)[:-1]) + (4, 4)
        T = np.zeros(shape, dtype=np.float32)
    if c is None:
        c = np.cos(x)
    if s is None:
        s = np.sin(x)

    T[..., 0, 0] = c
    T[..., 0, 1] = -s
    T[..., 1, 0] = s
    T[..., 1, 1] = c
    T[..., 2, 2] = 1

    return T

def invert(T, out=None):
    R = T[..., :3, :3]
    t = T[..., :3, 3:]

    if out is None:
        out = np.zeros_like(T)
    out[..., :3, :3] = R.swapaxes(-1, -2)
    out[..., :3, 3:] = -np.einsum('...ba,...bc->...ac', R, t)
    out[..., 3, 3] = 1
    return out


def Rti(R, t):
    Ri = R.swapaxes(-1, -2)
    if np.ndim(t) < np.ndim(Ri):
        # case (...,D)
        ti = -np.einsum('...ab,...b->...a', Ri, t)
    else:
        # case (...,D,1)
        ti = -np.einsum('...ab,...bc->...ac', Ri, t)
    return Ri, ti


def lerp(a, b, w):
    return (a * (1.0-w)) + (b*w)


def flerp(a, b, w, f, fi):
    return fi(lerp(f(a), f(b), w))


def rlerp(ra, rb, w):
    Ra = np.eye(4, dtype=np.float32)
    Rb = np.eye(4, dtype=np.float32)
    Ra[:3, :3] = ra
    Rb[:3, :3] = rb

    qa = tx.quaternion_from_matrix(Ra)
    qb = tx.quaternion_from_matrix(Rb)
    q = tx.quaternion_slerp(q0, q1, w)
    R = tx.quaternion_matrix(q)[:3, :3]
    return R

def rx3(R, x):
    rx = np.einsum('...ab,...b->...a', R[..., :3, :3], x)
    return rx

def tx3(T, x):
    rx = np.einsum('...ab,...b->...a', T[..., :3, :3], x)
    return rx + T[..., :3, 3:].swapaxes(-2, -1)


def rtx3(r, t, x):
    return x.dot(r.swapaxes(-2, -1)) + t


def tx4(T, x):
    return np.einsum('...ab,...b->...a', T, x)



import numpy as np
from cho_util.math.common import *


def from_matrix(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-2] + (3,))
    i, j, k = 0, 1, 2

    cy = np.sqrt(x[..., i, i]*x[..., i, i] + x[..., j, i]*x[..., j, i])

    msk = (cy > np.finfo(x.dtype).eps)
    x_a = x[msk]
    out[msk, 0] = np.arctan2(x_a[..., k, j], x_a[..., k, k])
    out[msk, 1] = np.arctan2(-x_a[..., k, i], cy[msk])
    out[msk, 2] = np.arctan2(x_a[..., j, i], x_a[..., i, i])

    nmsk = ~msk
    if nmsk.sum() > 0:
        raise ValueError("{}".format(nmsk.sum()))
    x_b = x[nmsk]
    out[nmsk, 0] = np.arctan2(-x_b[..., j, k],  x_b[..., j, j])
    out[nmsk, 1] = np.arctan2(-x_b[..., k, i],  cy[nmsk])
    out[nmsk, 2] = 0.0

    return out


def from_quaternion(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3,))

    qx, qy, qz, qw = [x[..., i] for i in range(4)]

    xx = 2.0 * qx*qx
    yy = 2.0 * qy*qy
    zz = 2.0 * qz*qz

    wx = 2.0 * qw*qx
    wy = 2.0 * qw*qy
    wz = 2.0 * qw*qz

    xy = 2.0 * qx*qy
    yz = 2.0 * qy*qz
    zx = 2.0 * qx*qz

    cy = np.sqrt(np.square(wz+xy) + np.square(yy+zz-1.0))
    out[..., 0] = np.arctan2(wx+yz, 1.0-xx-yy)
    out[..., 1] = np.arctan2(wy-zx, cy)
    out[..., 2] = np.arctan2(wz+xy, 1.0-yy-zz)

    return out


def from_euler(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3,))
    np.copyto(out, x)
    return out


def from_axis_angle(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3,))

    if x.shape[-1] == 3:
        # format : angle * axis
        angle = norm(x)
        axis = x / angle[..., None]
    elif x.shape[-1] == 4:
        # format : (axis, angle)
        axis = x[..., :3]
        angle = x[..., 3]
    else:
        raise ValueError("Invalid Input Shape : {}".format(x.shape))

    h = angle
    x, y, z = [axis[..., i] for i in range(3)]

    x0 = h
    x3 = np.cos(x0)
    x1 = np.sin(x0)
    x2 = y*z
    x10 = x*z
    x24 = x*y
    x4 = np.square(x)
    x7 = np.square(y)
    x16 = np.square(z)
    x11 = np.square(x7)
    x13 = np.square(x16)
    x5 = x3
    x6 = x4*x5
    x8 = x5*x7
    x9 = -x7 + x8 + 1
    x12 = x11
    x14 = x13
    x15 = 2.0*x7
    x17 = 2.0*x16
    x18 = x4*x7
    x19 = np.square(x5)
    x20 = x1
    x21 = x15*x16
    x22 = 2.0*x5
    x23 = x*x2*x20
    cy = np.sqrt(-x11*x22 + x12*x19 + x12 - x13*x22 + x14*x19 + x14 - x15*x6 - x15 + x16*np.square(x20) -
                 4.0*x16*x8 + x17*x5 - x17 + x18*x19 + x18 + x19*x21 + x21 - x22*x23 + 2.0*x23 + 2.0*x8 + 1.0)

    out[..., 0] = np.arctan2(x*x1 - x2*x3 + x2, -x4 + x6 + x9)
    out[..., 1] = np.arctan2(x1*y + x10*x3 - x10, cy)
    out[..., 2] = np.arctan2(x1*z - x24*x3 + x24, x16*x5 - x16 + x9)
    return out


def rotate(r, x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty_like(x)

    c = np.cos(r)
    s = np.sin(r)

    cx, cy, cz = [c[..., i] for i in range(3)]
    sx, sy, sz = [s[..., i] for i in range(3)]
    x, y, z = [x[..., i] for i in range(3)]

    x2 = -cx*sz
    x6 = cz*sx
    x7 = sz*sx
    x8 = cx*cz
    x10 = x*cy

    np.multiply(x10, cz, out=out[..., 0])
    out[..., 0] += y*(sy*x6+x2)
    out[..., 0] += z*(sy*x8+x7)

    np.multiply(sz, x10, out=out[..., 1])
    out[..., 1] += y*(sy*x7+x8)
    out[..., 1] -= z*(x2*sy+x6)

    np.multiply(-x, sy, out=out[..., 2])
    out[..., 2] += cx*cy*z
    out[..., 2] += sx*cy*y

    return out


def random(size=(), *args, **kwargs):
    size = tuple(np.reshape(size, [-1])) + (3,)
    return np.random.normal(size=size, *args, **kwargs)


def identity(dtype=np.float64):
    return np.zeros(3, dtype=dtype)



import argparse
from collections import defaultdict

try:
    import cv2
except ImportError as e:
    print('OpenCV does not exist, camera disabled : {}'.format(e))

class KeyCallback(object):
    def __init__(self):
        self.kmap_ = {
                255 : lambda: False,
                ord('q') : lambda: True
                }
    def log(self, k):
        print(k)
        return False
    def update(self, kcb):
        self.kmap_.update(kcb)
    def __call__(self, k):
        k = (k & 0xff)
        if k not in self.kmap_:
            print('k', k)
            return False
        f = self.kmap_[k]
        return f()


def main():
    # handle arguments
    parser = argparse.ArgumentParser(description='OpenCV Camera')
    parser.add_argument('--dev', dest='dev', type=str,
            help='camera device', required=True)

    args = parser.parse_args()
    cam = cv2.VideoCapture(args.dev)
    kcb = KeyCallback()

    while True:
        ret, img = cam.read()
        if not ret:
            print('ret', ret)
            break
        cv2.imshow('img', img)
        k = cv2.waitKey(1)
        if kcb(k):
            break


if __name__ == '__main__':
    main()



import numpy as np
from cho_util.math.common import *


def from_matrix(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-2] + (3, 3))
    np.copyto(out, x)
    return out


def from_quaternion(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3, 3))

    qx, qy, qz, qw = [x[..., i] for i in range(4)]

    tx = 2.0 * qx
    ty = 2.0 * qy
    tz = 2.0 * qz
    twx = tx * qw
    twy = ty * qw
    twz = tz * qw
    txx = tx * qx
    txy = ty * qx
    txz = tz * qx
    tyy = ty * qy
    tyz = tz * qy
    tzz = tz * qz

    out[..., 0, 0] = 1.0 - (tyy + tzz)
    out[..., 0, 1] = txy - twz
    out[..., 0, 2] = txz + twy

    out[..., 1, 0] = txy + twz
    out[..., 1, 1] = 1.0 - (txx + tzz)
    out[..., 1, 2] = tyz - twx

    out[..., 2, 0] = txz - twy
    out[..., 2, 1] = tyz + twx
    out[..., 2, 2] = 1.0 - (txx + tyy)

    return out


def from_euler(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3, 3))

    c, s = np.cos(x), np.sin(x)
    cx, cy, cz = [c[..., i] for i in range(3)]
    sx, sy, sz = [s[..., i] for i in range(3)]

    out[..., 0, 0] = cy * cz
    out[..., 0, 1] = (sx * sy * cz) - (cx * sz)
    out[..., 0, 2] = (cx * sy * cz) + (sx * sz)
    out[..., 1, 0] = cy * sz
    out[..., 1, 1] = (sx * sy * sz) + (cx * cz)
    out[..., 1, 2] = (cx * sy * sz) - (sx * cz)
    out[..., 2, 0] = -sy
    out[..., 2, 1] = sx * cy
    out[..., 2, 2] = cx * cy

    return out


def from_axis_angle(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (3, 3))

    if x.shape[-1] == 3:
        # format : angle * axis
        angle = norm(x, keepdims=True)
        axis = x / angle[..., None]
    elif x.shape[-1] == 4:
        # format : (axis, angle)
        axis = x[..., :3]
        angle = x[..., 3:]
    else:
        raise ValueError("Invalid Input Shape : {}".format(x.shape))

    sin_axis = np.sin(angle) * axis
    cos_angle = np.cos(angle)
    cos1_axis = (1.0 - cos_angle) * axis
    axis_y, axis_z = [axis[..., i] for i in (1, 2)]
    cos1_axis_x, cos1_axis_y = [cos1_axis[..., i] for i in (0, 1)]
    sin_axis_x, sin_axis_y, sin_axis_z = [sin_axis[..., i] for i in range(3)]
    tmp = cos1_axis_x * axis_y
    m01 = tmp - sin_axis_z
    m10 = tmp + sin_axis_z
    tmp = cos1_axis_x * axis_z
    m02 = tmp + sin_axis_y
    m20 = tmp - sin_axis_y
    tmp = cos1_axis_y * axis_z
    m12 = tmp - sin_axis_x
    m21 = tmp + sin_axis_x
    diag = cos1_axis * axis + cos_angle
    diag_x, diag_y, diag_z = [diag[..., i] for i in range(3)]

    out[..., 0, 0] = diag_x
    out[..., 0, 1] = m01
    out[..., 0, 2] = m02

    out[..., 1, 0] = m10
    out[..., 1, 1] = diag_y
    out[..., 1, 2] = m12

    out[..., 2, 0] = m20
    out[..., 2, 1] = m21
    out[..., 2, 2] = diag_z

    return out


def rotate(r, x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty_like(x)
    np.einsum('...ij,...j->...i', r, x, out=out)
    return out


def random(size=(), *args, **kwargs):
    # TODO(yycho0108): more efficient randomization
    size = tuple(np.reshape(size, [-1])) + (3, 3)
    out = np.random.normal(size=size, *args, **kwargs)
    out = [np.linalg.qr(o)[0] for o in out.reshape(-1, 3, 3)]
    out = np.reshape(out, size)
    return out


def inverse(r, out=None):
    if out is None:
        out = np.empty_like(r)
    np.copyto(out, r.swapaxes(-2, -1))
    return out

def identity(dtype=np.float64):
    return np.eye(3, dtype=dtype)



import numpy as np

from cho_util.math.rotation import _matrix
from cho_util.math.rotation import _quaternion
from cho_util.math.rotation import _euler
from cho_util.math.rotation import _axis_angle
from cho_util.math.rotation._euler import *


def to_matrix(x, *args, **kwargs):
    return _matrix.from_euler(x, *args, **kwargs)


def to_quaternion(x, *args, **kwargs):
    return _quaternion.from_euler(x, *args, **kwargs)


def to_euler(x, *args, **kwargs):
    return _euler.from_euler(x, *args, **kwargs)


def to_axis_angle(x, *args, **kwargs):
    return _axis_angle.from_euler(x, *args, **kwargs)


def inverse(r, out=None):
    r = np.asarray(r)
    if out is None:
        out = np.empty_like(r)
    return _euler.from_quaternion(_quaternion.inverse(_quaternion.from_euler(r)), out=out)

# def rotate(r, x, out=None):
#    # option : delegate to matrix
#    r = to_matrix(r)
#    out = _matrix.rotate(r, x, out)
#    return out



""" mostly matplotlib stuff """
import numpy as np

def set_axes_radius(ax, origin, radius):
    ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
    ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
    ax.set_zlim3d([origin[2] - radius, origin[2] + radius])

def set_axes_equal(ax):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    set_axes_radius(ax, origin, radius)




from cho_util.math.common import *
import numpy as np


def from_matrix(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-2] + (4,))

    m00, m01, m02 = [x[..., 0, i] for i in range(3)]
    m10, m11, m12 = [x[..., 1, i] for i in range(3)]
    m20, m21, m22 = [x[..., 2, i] for i in range(3)]

    np.subtract(m21, m12, out=out[..., 0])
    np.subtract(m02, m20, out=out[..., 1])
    np.subtract(m10, m01, out=out[..., 2])
    out[..., :3] = uvec(out[..., :3])
    out[..., 3] = np.arccos(np.clip((m00 + m11 + m22 - 1)*0.5, -1.0, 1.0))
    return out


def from_quaternion(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))
    qw = x[..., 3:]
    out[..., :3] = uvec(x[..., :3])
    out[..., 3:] = 2 * np.arccos(np.clip(qw, -1.0, 1.0))
    return out


def from_euler(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))

    x, y, z = [x[..., i] for i in range(3)]

    x0 = np.cos(x)
    x6 = np.sin(x)
    x3 = np.cos(y)
    x4 = np.sin(y)
    x5 = np.cos(z)
    x1 = np.sin(z)

    x2 = x0*x1
    x10 = x1*x6
    x7 = x5*x6
    x11 = x0*x5

    x8 = x1*x3 + x2 - x4*x7
    x9 = -x2*x4 + x3*x6 + x7
    x12 = x10 + x11*x4 + x4

    out[..., 0] = x9
    out[..., 1] = x12
    out[..., 2] = x8
    out[..., :3] = uvec(out[..., :3])

    out[..., 3] = np.arccos(
        np.clip(0.5 * (x0*x3 + x10*x4 + x11 + x3*x5 - 1), -1.0, 1.0))
    return out


def from_axis_angle(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))
    np.copyto(out, x)
    return out


def rotate(r, x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty_like(x)
    if r.shape[-1] == 3:
        # format : angle * axis
        angle = norm(r, keepdims=True)
        axis = r / angle
    elif r.shape[-1] == 4:
        # format : (axis, angle)
        axis = r[..., :3]
        angle = r[..., 3:]
    else:
        raise ValueError("Invalid Input Shape : {}".format(x.shape))
    u = axis
    c, s = np.cos(angle), np.sin(angle)
    d = (u*x).sum(axis=-1, keepdims=True)

    # out = (x*c) + s*np.cross(u, x) + (1.-c)*d*u
    np.multiply(x, c, out=out)
    out += s*np.cross(u, x)
    out += (1.-c)*d*u
    return out


def random(size=(), *args, **kwargs):
    size = tuple(np.reshape(size, [-1])) + (4,)
    out = np.random.normal(size=size, *args, **kwargs)
    out[..., :3] = uvec(out[..., :3])
    return out


def inverse(r, out=None):
    r = np.asarray(r)
    if out is None:
        out = np.empty_like(r)
    out[..., :3] = r[..., :3]
    out[..., 3] = -r[..., 3]
    return out


def identity(dtype=np.float64):
    return np.asarray([1, 0, 0, 0], dtype=dtype)



from unittest import TestCase
import numpy as np
import time

from cho_util.math.common import *
from cho_util.math import rotation

def random_point(size, *args, **kwargs):
    size = tuple(np.reshape(size, [-1])) + (3,)
    out = np.random.normal(size=size, *args, **kwargs)
    return out

def test_inverse():
    seed = np.random.randint(low=0,high=65535)
    print('seed  : {}'.format(seed))
    np.random.seed( seed )
    size = (1)
    n_iter = 100

    rtypes = {
            'm' : rotation.matrix,
            'q' : rotation.quaternion,
            #'e' : rotation.euler,
            'a' : rotation.axis_angle
            }
    source_set = list(rtypes.keys())

    seq = np.random.choice(len(source_set), size=n_iter, replace=True)
    seq = np.array(source_set)[seq]
    seq = ''.join(seq)

    for s in seq:
        r = rtypes[s].random(size=size, scale=1e-3)
        ri = rtypes[s].inverse(r)
        p = random_point(size)
        
        rir_p = rtypes[s].rotate(ri, rtypes[s].rotate(r, p))
        irr_p = rtypes[s].rotate(r, rtypes[s].rotate(ri, p))
        if not np.allclose(p, rir_p):
            print('rir_p failed for type : {}'.format(s))
        if not np.allclose(p, irr_p):
            print('irr_p failed for type : {}'.format(s))

def test_rotate():
    seed = np.random.randint(low=0,high=65535)
    print('seed  : {}'.format(seed))
    np.random.seed( seed )
    source_set = ['m', 'q', 'e', 'a']
    size = (1024)
    n_iter = 100
    seq_len = 2

    gen = {
        'm': rotation.matrix.random,
        'q': rotation.quaternion.random,
        'e': rotation.euler.random,
        'a': rotation.axis_angle.random,
    }

    r = {
            'm' : rotation.matrix.rotate,
            'q' : rotation.quaternion.rotate,
            'e' : rotation.euler.rotate,
            'a' : rotation.axis_angle.rotate
            }

    f = {
        'm,m': rotation.matrix.to_matrix,
        'm,q': rotation.matrix.to_quaternion,
        'm,e': rotation.matrix.to_euler,
        'm,a': rotation.matrix.to_axis_angle,

        'q,m': rotation.quaternion.to_matrix,
        'q,q': rotation.quaternion.to_quaternion,
        'q,e': rotation.quaternion.to_euler,
        'q,a': rotation.quaternion.to_axis_angle,

        'e,m': rotation.euler.to_matrix,
        'e,q': rotation.euler.to_quaternion,
        'e,e': rotation.euler.to_euler,
        'e,a': rotation.euler.to_axis_angle,

        'a,m': rotation.axis_angle.to_matrix,
        'a,q': rotation.axis_angle.to_quaternion,
        'a,e': rotation.axis_angle.to_euler,
        'a,a': rotation.axis_angle.to_axis_angle,
    }

    for _ in range(n_iter):
        ts = []
        ts.append(time.time())

        # random conversion sequence
        seq = np.random.choice(len(source_set), size=seq_len, replace=True)
        seq = np.array(source_set)[seq]
        seq = ''.join(seq)
        # uncomment below to debug
        #print(seq[:3] + '..' + seq[-3:])

        # initialization
        x = gen[seq[0]](size=size, scale=1e-3).astype(np.float32)
        point = random_point(size=size)

        R0 = f['{},{}'.format(seq[0], 'm')](x)
        initial_rotated_point = r[seq[0]](x, point)
        initial_rotated_point_m = r['m'](R0, point)
        #print('initial')
        #print(initial_rotated_point)

        for prv, nxt in zip(seq[:-1], seq[1:]):
            # print('{}->{}'.format(prv,nxt))
            # functions
            f_fw = f['{},{}'.format(prv, nxt)]
            f_bw = f['{},{}'.format(nxt, prv)]
            x = f_fw(x)
            if np.any(np.isnan(x)):
                print('x is nan : {}->{}'.format(prv, nxt))
                return

        R1 = f['{},{}'.format(seq[-1], 'm')](x)
        final_rotated_point = r[seq[-1]](x, point)
        final_rotated_point_m = r['m'](R1, point)
        #print('final'
        #print(final_rotated_point)

        error  = norm(final_rotated_point-initial_rotated_point)
        error_m  = norm(final_rotated_point_m-initial_rotated_point_m)
        Rerror = norm(R1-R0, axis=(-2,-1))
        if error.max() > 0.05:
            print('error (point) stats', error.min(), error.mean(), error.max())
            print('error (point-m) stats', error_m.min(), error_m.mean(), error_m.max())
            print('error (matrix) stats', Rerror.min(), Rerror.mean(), Rerror.max())
            ts.append(time.time())
            print('took {} sec'.format(np.diff(ts)))

def main():
    test_rotate()
    test_inverse()

if __name__ == '__main__':
    main()



try:
    from importlib.metadata import version, PackageNotFoundError
    try:
        __version__ = version("cho_util")
    except PackageNotFoundError:
        # package is not installed
        pass
except ModuleNotFoundError:
    # Python3.6 fallback
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        __version__ = get_distribution("cho_util").version
    except DistributionNotFound:
        # package is not installed
        pass

from . import cam
from . import math
from . import viz
from . import app



#!/usr/bin/env python2
import numpy as np
try:
    import cv2
except ImportError as e:
    print('OpenCV Import Error : {}'.format(e))

from cho_util import math as vm

def draw_lines(img1,img2,lines,pts1,pts2,cols,
        draw_pt=False
        ):
    ''' img1 - image on which we draw the epilines for the points in img2
        lines - corresponding epilines '''
    h,w = img1.shape[:2]
    for r,pt1,pt2,color in zip(lines,pts1,pts2,cols):
        x0,y0 = map(int, [0, -r[2]/r[1] ])
        x1,y1 = map(int, [w, -(r[2]+r[0]*w)/r[1] ])
        img1 = cv2.line(img1, (x0,y0), (x1,y1), color,1)
        if draw_pt:
            img1 = cv2.circle(img1, tuple(vm.rint(pt1)),5,color,-1)
            img2 = cv2.circle(img2, tuple(vm.rint(pt2)),5,color,-1)
    return img1,img2

def draw_points(img, pt,
        radius=None,
        color=None
        ):
    if radius is None:
        radius = vm.rint( 0.005 * np.min([img.shape[:2]]))
    pt = vm.rint(pt)
    for p in pt:
        cv2.circle(img, tuple(p), radius, color, radius)
    return img

def draw_matches(img1, img2, pt1, pt2,
        msk=None,
        radius=None,
        single=False
        ):
    if radius is None:
        radius = vm.rint( 0.005 * np.min([img1.shape[:2], img2.shape[:2]]) )
    h,w = np.shape(img1)[:2]

    if single:
        pt1 = np.round(pt1).astype(np.int32)
        pt2 = np.round(pt2).astype(np.int32)
        mim = img2.copy()
        mim0 = mim.copy()
    else:
        pt1 = np.round(pt1).astype(np.int32)
        pt2 = np.round(pt2 + [[w,0]]).astype(np.int32)
        mim = np.concatenate([img1, img2], axis=1)
        mim[:,w ] = 0
        mim0 = mim.copy()

    if msk is None:
        msk = np.ones(len(pt1), dtype=np.bool)

    n = msk.sum()
    col = np.random.randint(255, size=(n,3))

    for (p1, p2, c) in zip(pt1[msk], pt2[msk], col):
        p1 = tuple(p1)
        p2 = tuple(p2)
        cc = tuple([int(e) for e in vm.rint(c)])
        cv2.line(mim, p1, p2, cc, radius)
    mim = cv2.addWeighted(mim0, 0.5, mim, 0.5, 0.0)

    for (p1, p2, c) in zip(pt1[msk], pt2[msk], col):
        cc = tuple([int(e) for e in vm.rint(c)])
        #cc = (255,255,255)
        cv2.circle(mim, tuple(p1), radius, cc, -1)
        cv2.circle(mim, tuple(p2), radius, cc, -1)

    for p in pt1[~msk]:
        cv2.circle(mim, tuple(p), radius, (255,0,0), -1)

    for p in pt2[~msk]:
        cv2.circle(mim, tuple(p), radius, (255,0,0), -1)
    return mim

def make_color_wheel():
    """
    Generate color wheel according Middlebury color code
    :return: Color wheel
    """
    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR

    colorwheel = np.zeros([ncols, 3])

    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.transpose(np.floor(255*np.arange(0, RY) / RY))
    col += RY

    # YG
    colorwheel[col:col+YG, 0] = 255 - np.transpose(np.floor(255*np.arange(0, YG) / YG))
    colorwheel[col:col+YG, 1] = 255
    col += YG

    # GC
    colorwheel[col:col+GC, 1] = 255
    colorwheel[col:col+GC, 2] = np.transpose(np.floor(255*np.arange(0, GC) / GC))
    col += GC

    # CB
    colorwheel[col:col+CB, 1] = 255 - np.transpose(np.floor(255*np.arange(0, CB) / CB))
    colorwheel[col:col+CB, 2] = 255
    col += CB

    # BM
    colorwheel[col:col+BM, 2] = 255
    colorwheel[col:col+BM, 0] = np.transpose(np.floor(255*np.arange(0, BM) / BM))
    col += + BM

    # MR
    colorwheel[col:col+MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
    colorwheel[col:col+MR, 0] = 255

    return colorwheel

def compute_color(u, v):
    """
    compute optical flow color map
    :param u: optical flow horizontal map
    :param v: optical flow vertical map
    :return: optical flow in color code
    """
    [h, w] = u.shape
    img = np.zeros([h, w, 3])
    nanIdx = np.isnan(u) | np.isnan(v)
    u[nanIdx] = 0
    v[nanIdx] = 0

    colorwheel = make_color_wheel()
    ncols = np.size(colorwheel, 0)

    rad = np.sqrt(u**2+v**2)

    a = np.arctan2(-v, -u) / np.pi

    fk = (a+1) / 2 * (ncols - 1) + 1

    k0 = np.floor(fk).astype(int)

    k1 = k0 + 1
    k1[k1 == ncols+1] = 1
    f = fk - k0

    for i in range(0, np.size(colorwheel,1)):
        tmp = colorwheel[:, i]
        col0 = tmp[k0-1] / 255
        col1 = tmp[k1-1] / 255
        col = (1-f) * col0 + f * col1

        idx = rad <= 1
        col[idx] = 1-rad[idx]*(1-col[idx])
        notidx = np.logical_not(idx)

        col[notidx] *= 0.75
        img[:, :, i] = np.uint8(np.floor(255 * col*(1-nanIdx)))

    return img

def flow_to_image(flow, display=False, thresh=1e7):
    """
    Convert flow into middlebury color code image
    :param flow: optical flow map
    :return: optical flow image in middlebury color
    """
    # from https://github.com/vt-vl-lab/tf_flownet2.git
    u = flow[:, :, 0]
    v = flow[:, :, 1]

    maxu = -999.
    maxv = -999.
    minu = 999.
    minv = 999.

    idxUnknow = (abs(u) > thresh) | (abs(v) > thresh)
    u[idxUnknow] = 0
    v[idxUnknow] = 0

    maxu = max(maxu, np.max(u))
    minu = min(minu, np.min(u))

    maxv = max(maxv, np.max(v))
    minv = min(minv, np.min(v))

    rad = np.sqrt(u ** 2 + v ** 2)
    maxrad = max(-1, np.max(rad))

    if display:
        print("max flow: %.4f\nflow range:\nu = %.3f .. %.3f\nv = %.3f .. %.3f" % (maxrad, minu,maxu, minv, maxv))

    u = u/(maxrad + np.finfo(float).eps)
    v = v/(maxrad + np.finfo(float).eps)

    img = compute_color(u, v)

    idx = np.repeat(idxUnknow[:, :, np.newaxis], 3, axis=2)
    img[idx] = 0

    return np.uint8(img)




import numpy as np

inv = np.linalg.inv


def rint(x):
    """ convert to rounded integer (int32) """
    return np.round(x).astype(np.int32)


def anorm(x):
    """ angular norm, convert to range [-pi,pi] """
    return (x + np.pi) % (2*np.pi) - np.pi


def norm(x, *args, **kwargs):
    if 'axis' not in kwargs:
        kwargs['axis'] = -1
    return np.linalg.norm(x, *args, **kwargs)


def uvec(x):
    """ convert to unit vector """
    n = norm(x, keepdims=True)
    return np.divide(x, n, out=np.zeros_like(x),
                     where=(n > np.finfo(x.dtype).eps))


def lerp(a, b, w):
    """ linear interpolation """
    return (a * (1.0-w)) + (b*w)


__all__ = ['rint', 'anorm', 'norm', 'uvec', 'lerp', 'inv']



import numpy as np

from cho_util.math.rotation import _matrix
from cho_util.math.rotation import _quaternion
from cho_util.math.rotation import _euler
from cho_util.math.rotation import _axis_angle
from cho_util.math.rotation._matrix import *


def to_matrix(x, *args, **kwargs):
    return _matrix.from_matrix(x, *args, **kwargs)


def to_quaternion(x, *args, **kwargs):
    return _quaternion.from_matrix(x, *args, **kwargs)


def to_euler(x, *args, **kwargs):
    return _euler.from_matrix(x, *args, **kwargs)


def to_axis_angle(x, *args, **kwargs):
    return _axis_angle.from_matrix(x, *args, **kwargs)



#!/usr/bin/env python3

import pytest

from dataclasses import dataclass
from cho_util.app.with_cli import with_cli


@dataclass
class AppConfig:
    num_iter: int = 128


def inner_main(cfg: AppConfig):
    return cfg.num_iter

def invalid_inner_main(cfg):
    return cfg.num_iter

def test_cli():
    assert with_cli(inner_main)() == 128
    assert with_cli(AppConfig)(inner_main)() == 128
    assert with_cli(AppConfig(num_iter=64))(inner_main)() == 64
    assert with_cli(argv=['--num_iter', '32'])(inner_main)() == 32
    assert with_cli(argv=[])(inner_main)() == 128

    with_cli(AppConfig)(invalid_inner_main)()
    with_cli(cls=AppConfig)(invalid_inner_main)()



if __name__ == '__main__':
    test_cli()



#!/usr/bin/env python3

import os
import sys
import json
import yaml

from functools import partial
from collections import OrderedDict
from pathlib import Path


def file_cache(name_fn, load_fn, dump_fn, binary: bool = True):
    """Decorator for caching a result from a function to a file.

    The caller is responsible for providing a `name_fn` arg that
    determines a unique key for a given function.
    """
    def call_or_load(compute_fn):
        def wrapper(*args, **kwargs):
            filename = name_fn(compute_fn, *args, **kwargs)
            cache_file = Path(filename)
            # Compute if non-existent.
            if not cache_file.exists():
                # Ensure directory exists.
                cache_file.parent.mkdir(parents=True, exist_ok=True)
                result = compute_fn(*args, **kwargs)
                mode = 'wb' if binary else 'w'
                with open(filename, mode) as f:
                    dump_fn(result, f)
                return result

            # O.W. Return from cache.
            mode = 'rb' if binary else 'r'
            with open(filename, mode) as f:
                return load_fn(f)

        return wrapper
    return call_or_load


try:
    import json
    json_load = partial(json.load, object_pairs_hook=OrderedDict)
    json_file_cache = partial(file_cache,
                              load_fn=json.load,
                              dump_fn=json.dump,
                              binary=False)
except ImportError:
    pass

try:
    import yaml
    yaml_file_cache = partial(file_cache,
                              load_fn=yaml.safe_load,
                              dump_fn=yaml.safe_dump,
                              binary=False)
except ImportError:
    pass

try:
    import pickle
    pickle_file_cache = partial(file_cache,
                                load_fn=pickle.load,
                                dump_fn=pickle.dump,
                                binary=True)
except ImportError:
    pass



#!/usr/bin/env python2

import numpy as np
import cho_util.math.rotation as rotation

def print_Rt(R, t, round=2):
    print('\tR', np.round(np.rad2deg(rotation.euler.from_matrix(R)), round))

    n = np.linalg.norm(t)
    u = (t.ravel()/n  if n > np.finfo(np.float32).eps else t.ravel())
    print('\tt', np.round(u, round))

def print_ratio(a, b, name=''):
    q = (np.nan if (np.isscalar(b) and np.isclose(b, 0)) else (100.0*a)/b)
    print('[{}] {}/{}={}%'.format(name, a, b, q))



#!/usr/bin/env python3

# import numba as nb
import numpy as np
from cho_util.math.common import *

I_X = 0
I_Y = 1
I_Z = 2
I_W = 3


def from_matrix(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-2] + (4,))

    # parse input
    m00, m01, m02 = [x[..., 0, i] for i in range(3)]
    m10, m11, m12 = [x[..., 1, i] for i in range(3)]
    m20, m21, m22 = [x[..., 2, i] for i in range(3)]

    out[..., I_X] = 1. + m00 - m11 - m22
    out[..., I_Y] = 1. - m00 + m11 - m22
    out[..., I_Z] = 1. - m00 - m11 + m22
    out[..., I_W] = 1. + m00 + m11 + m22

    # sqrt(max(0,x))/2
    for i in range(4):
        np.maximum(0.0, out[..., i], out=out[..., i])
    np.sqrt(out, out=out)
    out *= 0.5

    # correct for signs
    np.copysign(out[..., I_X], m21-m12, out=out[..., I_X])
    np.copysign(out[..., I_Y], m02-m20, out=out[..., I_Y])
    np.copysign(out[..., I_Z], m10-m01, out=out[..., I_Z])

    return out


def from_quaternion(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))
    np.copyto(out, x)
    return out


def from_euler(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))

    half_x = (0.5 * x)
    c, s = np.cos(half_x), np.sin(half_x)
    cx, cy, cz = [c[..., i] for i in range(3)]
    sx, sy, sz = [s[..., i] for i in range(3)]

    cc = cx*cz
    cs = cx*sz
    sc = sx*cz
    ss = sx*sz

    out[..., 0] = cy*sc - sy*cs
    out[..., 1] = cy*ss + sy*cc
    out[..., 2] = cy*cs - sy*sc
    out[..., 3] = cy*cc + sy*ss

    return out


def from_axis_angle(x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty(shape=np.shape(x)[:-1] + (4,))

    if x.shape[-1] == 3:
        # format : angle * axis
        angle = norm(x, keepdims=True)
        axis = x / angle
    elif x.shape[-1] == 4:
        # format : (axis, angle)
        axis = x[..., :3]
        angle = x[..., 3:]
    else:
        raise ValueError("Invalid Input Shape : {}".format(x.shape))
    half_angle = 0.5 * angle
    out[..., :3] = np.sin(half_angle) * axis
    out[..., 3:] = np.cos(half_angle)
    return out


# @nb.njit
def multiply(q1, q2, out=None):
    q1 = np.asarray(q1)
    q2 = np.asarray(q2)
    if out is None:
        out = np.empty_like(q1)
    x1, y1, z1, w1 = [q1[..., i] for i in range(4)]
    x2, y2, z2, w2 = [q2[..., i] for i in range(4)]
    out[..., 0] = x1 * w2 + y1 * z2 - z1 * y2 + w1 * x2
    out[..., 1] = -x1 * z2 + y1 * w2 + z1 * x2 + w1 * y2
    out[..., 2] = x1 * y2 - y1 * x2 + z1 * w2 + w1 * z2
    out[..., 3] = -x1 * x2 - y1 * y2 - z1 * z2 + w1 * w2
    return out


# @nb.njit
def rotate(r, x, out=None):
    x = np.asarray(x)
    if out is None:
        out = np.empty_like(x)

    # Slower?
    #t = 2.0 * np.cross(r[..., :3], x)
    #out[:] = x + r[..., -1:] * t + np.cross(r[..., :3], t)

    qx, qy, qz, qw = [r[..., i] for i in range(4)]
    vx, vy, vz = [x[..., i] for i in range(3)]

    x0 = qw*vx + qy*vz - qz*vy
    x1 = qx*vx + qy*vy + qz*vz
    x2 = qw*vz + qx*vy - qy*vx
    x3 = qw*vy - qx*vz + qz*vx
    out[..., 0] = qw*x0 + qx*x1 + qy*x2 - qz*x3
    out[..., 1] = qw*x3 - qx*x2 + qy*x1 + qz*x0
    out[..., 2] = qw*x2 + qx*x3 - qy*x0 + qz*x1
    return out


def random(size=(), *args, **kwargs):
    size = tuple(np.reshape(size, [-1])) + (4,)
    out = np.random.normal(size=size, *args, **kwargs)
    out = uvec(out)
    return out


def conjugate(q, out=None):
    q = np.asarray(q)
    if out is None:
        out = np.empty_like(q)
    out[..., :3] = -q[..., :3]
    out[..., 3] = q[..., 3]
    return out


def inverse(q, out=None):
    # NOTE: assume unit quaternion
    return conjugate(q, out)


def identity(dtype=np.float64):
    return np.asarray([0, 0, 0, 1], dtype=dtype)



#!/usr/bin/env python3
#PYTHON_ARGCOMPLETE_OK
"""Automatically create argparse parser for CLI arguments from Serializable
dataclass."""

__all__ = ['with_cli']

import sys
from dataclasses import dataclass, is_dataclass
from simple_parsing import ArgumentParser, Serializable
from typing import Callable, List, Type, TypeVar, Union, Optional
import logging
import argparse
import inspect
from pathlib import Path
from functools import wraps

use_argcomplete = False
try:
    import argcomplete
    use_argcomplete = True
except ImportError:
    logging.info('argcomplete disabled due to missing package.')


D = TypeVar("D")


def _update_settings_from_file(opts: dataclass, argv: List[str]):
    """Update given settings from a configuration file."""
    config_parser = argparse.ArgumentParser(add_help=False)
    config_parser.add_argument(
        '--load_config',
        help='Load from configuration file',
        dest='load_config',
        metavar='CONFIG_FILE',
        default='')
    parsed_args, next_argv = config_parser.parse_known_args(argv)

    # If no `config_file` arg, fallback to passthrough behavior.
    if not parsed_args.load_config:
        return (opts, argv, [config_parser])

    config_file = Path(parsed_args.load_config)
    if not config_file.exists():
        logging.error(F'config file = {config_file} does not exist!')
        return (opts, argv, [config_parser])

    opts = opts.load(str(config_file))
    return (opts, next_argv, [config_parser])


def update_settings(opts: dataclass,
                    argv: Optional[List[str]] = None,
                    description: Optional[str] = None):
    """Update given settings from command line arguments.

    Uses `argparse`, `argcomplete` and `simple_parsing` under the hood.
    """
    if not is_dataclass(opts):
        raise ValueError(F'Cannot update args on non-dataclass class : {opts}')
    cls = opts if isinstance(opts, type) else type(opts)

    # Use default system argv if not supplied.
    argv = sys.argv[1:] if argv is None else argv

    # Update from config file, if applicable.
    parser_parents = []
    if issubclass(cls, Serializable):
       opts, argv, parser_parents = _update_settings_from_file(opts, argv)

    # Update from cli args...
    parser = ArgumentParser(parents=parser_parents, description=description)
    if isinstance(opts, type):
        parser.add_arguments(cls, dest='opts')
    else:
        parser.add_arguments(cls, dest='opts', default=opts)
    if use_argcomplete:
        # NOTE(ycho): required to call `_preprocessing()` prior to autocomplete
        # due to some arguments being unfilled up to that point in time.
        parser._preprocessing()
        argcomplete.autocomplete(parser)
    args, unargs = parser.parse_known_args(argv, attempt_to_reorder=True)
    if len(unargs) > 0:
        logging.warn(F'Ignored args = {unargs}')
    return args.opts


def with_cli(cls: Union[Type[D], D] = None, argv: List[str] = None):
    """Decorator for automatically adding parsed args from cli to entry
    point."""

    main = None
    if cls is None:
        # @with_cli()
        need_cls = True
    else:
        if callable(cls) and not is_dataclass(cls):
            # @with_cli
            main = cls
            need_cls = True
        else:
            # @with_cli(cls=Config, ...)
            need_cls = (cls is None)  # FIXME(ycho): always False.

    def decorator(main: Callable[[D], None]):
        # NOTE(ycho):
        # if `cls` is None, try to infer them from `main` signature.
        inner_cls = cls
        if need_cls:
            sig = inspect.signature(main)
            if len(sig.parameters) == 1:
                key = next(iter(sig.parameters))
                inner_cls = sig.parameters[key].annotation
            else:
                raise ValueError(
                    '#arg != 1 in main {}: Cannot infer param type.'
                    .format(sig))

        # If supplied, load from file.
        # The arguments provided through the file will be
        # overridden by any CLI args if present.
        instance = None

        # NOTE(ycho): using @wraps to forward main() documentation.
        @wraps(main)
        def wrapper():
            doc = getattr(main, '__doc__', '')
            opts: D = update_settings(
                inner_cls, argv, description=doc)
            return main(opts)
        return wrapper

    if main is not None:
        return decorator(main)
    else:
        return decorator


def main():
    logging.basicConfig(level=logging.INFO)

    @dataclass
    class Settings:
        value: int = 0

    opts = Settings()
    opts = update_settings(opts)
    logging.info(F'got value = {opts.value}')


@dataclass
class AppConfig(Serializable):
    value: int = 1


@with_cli
def main2(cfg: AppConfig):
    """ @decorator """
    print(cfg)


@with_cli(AppConfig)
def main3(cfg: AppConfig):
    """ @decorator(...) """
    print(cfg)


@with_cli()
def main4(cfg: AppConfig):
    """ @decorator() """
    print(cfg)


@with_cli(AppConfig)
def main5(cfg):
    """ @decorator(cls), without function level type annotation """
    print(cfg)


@with_cli(AppConfig(value=16))
def main6(cfg):
    """ @decorator(cls(*args,**kwds)), i.e. called with instance"""
    print(cfg)


def main7():
    """test all the things."""
    def _main(cfg: AppConfig):
        """type-annotated main()"""
        print(cfg)

    def _main2(cfg):
        """not type-annotated main()"""
        print(cfg)

    # Null case
    _main(AppConfig())
    # case A
    with_cli()(_main)()
    # case B
    with_cli(_main)()
    # case C
    with_cli(AppConfig)(_main)()
    # case C
    with_cli(AppConfig(value=32))(_main)()
    # case C
    with_cli(argv=[])(_main)()
    # case C
    with_cli(argv=['--value', '64'])(_main)()

    # Tests for cases where arg is not type-anntoated:
    # INVALID: case A (cannot infer type)
    # with_cli()(_main2)()
    # INVALID: case B (cannot infer type)
    # with_cli(_main2)()
    # case C
    with_cli(AppConfig)(_main2)()
    with_cli(cls=AppConfig)(_main2)()


if __name__ == '__main__':
    main()
    print(main.__doc__)
    main2()
    print(main2.__doc__)
    main3()
    print(main3.__doc__)
    main4()
    print(main4.__doc__)
    main5()
    print(main5.__doc__)
    main6()
    print(main6.__doc__)
    main7()
    print(main7.__doc__)



#!/usr/bin/env python3

import re
from enum import Enum
from simple_parsing.helpers.serialization import (encode, register_decoding_fn)
from simple_parsing import Serializable, choice
from functools import partial
from typing import Tuple, Union, Iterable, List
from dataclasses import (
    dataclass,
    is_dataclass,
    make_dataclass,
    field)


def _encode_enum(o: 'EnumBase'):
    return o.name


def _decode_enum(cls: 'EnumBase', name: str):
    return getattr(cls, name)


def _encode_one_of(o):
    return encode({'which': o.which, 'value': o.value})


def _decode_one_of(cls, data):
    inst = cls(which=data['which'])
    inst.value = data['value']


class EnumBase(Enum):
    def __init_subclass__(cls):
        super().__init_subclass__()
        encode.register(cls, _encode_enum)
        register_decoding_fn(cls, partial(_decode_enum, cls))


def _clean(s: str):
    """cleanup invalid characters in `s`"""
    # Replace invalid characters with _.
    s = re.sub('[^0-9a-zA-Z_]', '_', s)
    # Remove leading characters until we find a letter or underscore
    s = re.sub('^[^a-zA-Z_]+', '', s)
    return s.lower()


def _unique(seq: Iterable) -> List:
    """order-preserving unique elements in `seq`."""
    out = []
    seen = set()
    for x in seq:
        if x in seen:
            continue
        out.append(x)
        seen.add(x)
    return out


def one_of(*clss: Tuple[Union[dataclass, Tuple[dataclass, str]], ...],
           typename: str = None) -> dataclass:
    """Creates a variant(cls...) where a particular option is selectable
    amongst many."""

    if len(clss) <= 0:
        raise IndexError(F'len(clss) must be > 0!')

    # Reduce to single `cls`
    if len(clss) == 1:
        cls = clss[0]
        if is_dataclass(cls):
            return cls

    # NOTE(ycho): Handle case where `typename` was not given as a kwd.
    if isinstance(clss[-1], str) and typename is None:
        clss = clss[:-1]
        typename = clss[-1]

    # Account for Tuple[dataclass,str]
    def _sanitize(clss):
        out_clss = []
        out_names = {}
        defaults = {}

        for cls in clss:
            try:
                cls, name = cls
                if not isinstance(cls, type):
                    inst = cls
                    cls = type(inst)
                    defaults[cls] = inst
                out_clss.append(cls)
                out_names[cls] = name
            except TypeError:
                if not isinstance(cls, type):
                    inst = cls
                    cls = type(inst)
                    defaults[cls] = inst
                out_clss.append(cls)
        return out_clss, out_names, defaults

    clss, cls_name, defaults = _sanitize(clss)

    # Make unique ...
    clss = tuple(_unique(clss))

    # Generate mapping between class and string-based valid identifier.
    for cls in clss:
        if cls in cls_name:
            continue
        cls_name[cls] = _clean(cls.__qualname__)

    if typename is None:
        typename = 'one_of_' + '_'.join(cls_name[cls] for cls in clss)
        typename = _clean(typename)

    Which = EnumBase('Which', {cls_name[cls]: i
                               for i, cls in enumerate(clss)})
    fields = [
        ('value', Union[clss], field(init=False)),
        # FIXME(ycho): choice doesn't appear to work ... for now.
        # ('which', Which, choice(Which, default=Which(0))),
        ('which', Which, Which(0))
    ]
    for cls in clss:
        default = None
        # Resolve defaults either from
        # passed-in instance or nullary-arg initialization.
        if cls in defaults:
            default = defaults[cls]
        else:
            try:
                default = cls()
            except TypeError:
                default = None

        if default is None:
            fields.append((cls_name[cls], cls))
        else:
            fields.append(
                (cls_name[cls], cls, field(
                    default=default)))

    def _set_value(self):
        name = cls_name[clss[self.which.value]]
        self.value = getattr(self, name)

    def _getattr(self, name):
        """propagate self.value.{name}"""
        return getattr(self.value, name)

    cls = make_dataclass(typename, fields,
                         bases=(Serializable,),
                         namespace={
                             '__post_init__': _set_value,
                             '__getattr__': _getattr,
                         })

    # encode.register(cls, _encode_one_of)
    # register_decoding_fn(cls, partial(_decode_one_of, cls))
    return cls
